(**************************************************************************)
(* Comb: sample code to show dynamic choice of combinator semantics       *)
(*                                                                        *)
(*  Author(s):  Roberto Di Cosmo                                          *)
(*                                                                        *)
(*  This library is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU Lesser General Public License as        *)
(*  published by the Free Software Foundation, either version 2 of the    *)
(*  License, or (at your option) any later version.  A special linking    *)
(*  exception to the GNU Lesser General Public License applies to this    *)
(*  library, see the LICENSE file for more information.                   *)
(**************************************************************************)

(* needs Functory, from functory.lri.fr *)
(* with opam, just type                 *)
(*         opam install functory        *)

(* compile with 
   ocamlfind ocamlopt -linkpkg -package functory,graphics -o comb.native comb.ml
 *)

(** The GADT defining List combinators *)

type ('a,'b) lcomb =
| Map : ('a -> 'b) -> ('a list,'b list) lcomb
| Fold_left : ('a -> 'b -> 'a) * 'a -> ('b list,'a) lcomb
| Fold_right : ('a -> 'b -> 'b) * 'b -> ('a list,'b) lcomb
| Comb : ('a,'b) lcomb * ('b,'c) lcomb -> ('a,'c) lcomb

(* Executing combinators *)

let rec exec : type a b. a -> (a,b) lcomb -> b = 
fun l -> 
 function
 | Map f -> List.map f l (* here 'a = 'c list and 'b = 'd list*)
 | Fold_left (f,e) -> List.fold_left f e l
 | Fold_right (f,e) -> List.fold_right f l e
 | Comb (c1,c2) -> let l' = exec l c1 in exec l' c2

(* Example of an optimization function *)

let rec optimize : type a b. (a,b) lcomb -> (a,b) lcomb =
function
| Comb (c1,c2) -> 
   (match optimize c1, optimize c2 with
    | Map f, Map g -> Map (fun x -> g (f x))
    | c,c' -> Comb (c,c'))
| x -> x

(* Drawing combinator sequences *)

open Graphics

let rec draw_step name x0 y0 x1 y1 =
  let xc = (x1-x0)/2 and yc = (y1-y0)/2 in
  let r  = truncate((float (min xc yc))*.0.80) in
  let dr = truncate((float r) *. 0.20) in
  let lname = (fst (text_size name)) in
    draw_circle (xc+x0) (y0+yc) r ;
    moveto x0 (y0+(y1-y0)/2); lineto (x0+dr) (y0+(y1-y0)/2);
    moveto (x1-dr) (y0+(y1-y0)/2); lineto (x1)  (y0+(y1-y0)/2);
    moveto (xc+x0-(lname/2)) (y0+yc); draw_string name

and draw_comb 
  : type a b c.(a,b) lcomb * (b,c) lcomb -> int -> int -> int -> int -> unit 
  = fun (c1,c2) x0 y0 x1 y1 ->
  let dx = (x1-x0)/2 in
  draw_expr c1 x0 y0 (x0+dx) y1;
  draw_expr c2 (x0+dx) y0 (x0+2*dx) y1;

and draw_expr 
  : type a b.(a,b) lcomb -> int -> int -> int -> int -> unit
  = fun e x0 y0 x1 y1 ->
  match e with
    Map s        -> (draw_step "map" x0 y0 x1 y1) 
  | Fold_left (_,_)  -> (draw_step "fold_left" x0 y0 x1 y1) 
  | Fold_right (_,_) -> (draw_step "fold_right" x0 y0 x1 y1) 
  | Comb (c1,c2) -> (draw_comb (c1,c2) x0 y0 x1 y1)

(* The interface of a list combinator implementation *)

module type COMB = 
sig
  type ('a,'b) t
  val map : ('a -> 'b) -> ('a list, 'b list) t
  val fold_left : ('a -> 'b -> 'a) -> 'a -> ('b list, 'a) t
  val fold_right : ('a -> 'b -> 'b) -> 'b -> ('a list, 'b) t
  val ( |> ) : ('a, 'b) t -> ('b, 'c) t -> ('a, 'c) t
  val optimize : ('a, 'b) t -> ('a, 'b) t
  val ( @ ) : ('a, 'b) t -> 'a -> 'b
end

(** List combinator implementation using the List from OCaml *)

module CombPlain : COMB= 
struct
 type ('a,'b) t = 'a -> 'b

 let map f l = List.map f l
 let fold_left f e l = List.fold_left f e l
 let fold_right f e l = List.fold_right f l e

 let ( |> ) = fun f g l -> g (f l)

 let rec optimize c = c

 let ( @ ) c l = c l
end;;

(** List combinator implementation using exec *)

module CombExecInternal = 
struct
 (* definition des combinateurs avec @ = exec *) 
 type ('a,'b) t = ('a,'b) lcomb
 let map f = Map f 
 let fold_left f e = Fold_left (f,e)
 let fold_right f e = Fold_right (f,e)

 let ( |> ) = fun c c' -> Comb (c,c')

 let optimize = optimize

 let ( @ ) c l = exec l c
end;;

(** List combinator implementation using draw *)

module CombDraw : COMB = 
struct
 include CombExecInternal
 let ( @ ) c l = 
    open_graph " 1024x800"; 
    draw_expr c 0 0 (size_x()) (size_y());
    print_string "Enter a newline to stop ...\n"; 
    let _ = read_line() in
    close_graph(); exec l c
end;;

(** List combinator implementation using Parmap *)

module CombPar : COMB = 
struct
 include CombExecInternal
 let rec exec : type a b. a -> (a,b) lcomb -> b = 
 fun l -> 
  function
  | Map f -> Functory.Cores.map ~f l
  | Fold_left (f,e) ->  Functory.Cores.map_local_fold ~f:(fun x -> x) ~fold:f e l
  | Fold_right (f,e) -> Functory.Cores.map_local_fold ~f:(fun x -> x) ~fold:(fun x y -> f y x) e l
  | Comb (Map f,Fold_left(g,e)) -> Functory.Cores.map_local_fold ~f:f ~fold:g e l
  | Comb (Map f,Fold_right(g,e)) -> Functory.Cores.map_local_fold ~f ~fold:(fun x y -> g y x) e l
  | Comb (c1,c2) -> let l' = exec l c1 in exec l' c2

 let ( @ ) c l = exec l c
end;;

(* restrict CombExec after defining CombDraw *)
(* to share the abtract type definition *)

module CombExec = (CombExecInternal : COMB);;

(* parse the command line options *)

let mode=ref (`Exec : [< `Exec | `Seq | `Gra | `Par of int]) ;;
let set m = mode := m;;

Arg.parse [
   ("-draw", Arg.Unit (fun () -> set `Gra), "Draw combinator sequences.");
   ("-plain", Arg.Unit (fun () -> set `Seq), "Use directly the list combinators.");
   ("-par", Arg.Int (fun n -> set (`Par n);
                     Functory.Control.set_debug true;
                     Functory.Cores.set_number_of_cores n), 
            "Use Functory to execute on n cores.");
  ] (fun _ -> ()) "Test combinators: try me with and without -draw";;

(* choose the right implementation, using first class modules *)

module Comb = 
  (val 
      (match !mode with
        `Gra -> (module CombDraw : COMB) 
      | `Exec -> (module CombExec : COMB)
      | `Seq -> (module CombPlain : COMB)
      | `Par n -> (module CombPar : COMB)
      ) : COMB);;

(* and use it in the rest of the program *)

open Comb;;

let myop = map (fun x -> x+1) |> map (fun x -> x*2);;  
let pr_list l = List.iter (fun n -> Printf.printf "%d " n) l; Printf.printf "\n%!";;

pr_list (myop @ [1;2;3]);;

pr_list((optimize myop) @ [1;2;3]);;
